# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: Copyright (c) 2018 raimon
# SPDX-FileCopyrightText: Copyright (c) 2025 stefan6419846


class PipLicensesWarning(UserWarning):
    """
    Base class for warnings emitted by the pip-licenses-cli package.
    """



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: Copyright (c) 2018 raimon
# SPDX-FileCopyrightText: Copyright (c) 2025 stefan6419846

import sys

from piplicenses import __pkgname__

FIELD_NAMES = (
    "Name",
    "Version",
    "License",
    "LicenseFile",
    "LicenseText",
    "NoticeFile",
    "NoticeText",
    "Author",
    "Maintainer",
    "Description",
    "URL",
)


SUMMARY_FIELD_NAMES = (
    "Count",
    "License",
)


DEFAULT_OUTPUT_FIELDS = (
    "Name",
    "Version",
)


SUMMARY_OUTPUT_FIELDS = (
    "Count",
    "License",
)

# Mapping of FIELD_NAMES to METADATA_KEYS where they differ by more than case
FIELDS_TO_METADATA_KEYS = {
    "URL": "homepage",
    "License-Metadata": "license",
    "License-Classifier": "license_classifier",
    "LicenseFile": "license_files",
    "LicenseFiles": "license_files",
    "LicenseText": "license_texts",
    "LicenseTexts": "license_texts",
    "NoticeFile": "notice_files",
    "NoticeFiles": "notice_files",
    "NoticeText": "notice_texts",
    "NoticeTexts": "notice_texts",
    "OtherFiles": "other_files",
    "OtherTexts": "other_texts",
    "Description": "summary",
}

_MULTI_VALUE_KEYS = {
    "LicenseFile",
    "LicenseFiles",
    "LicenseText",
    "LicenseTexts",
    "NoticeFile",
    "NoticeFiles",
    "NoticeText",
    "NoticeTexts",
    "OtherFiles",
    "OtherTexts",
}


SYSTEM_PACKAGES = [
    __pkgname__,
    "pip",
    "pip-licenses-lib",
    "prettytable",
    "wcwidth",
    "setuptools",
    "wheel",
]
if sys.version_info < (3, 11):  # pragma: no cover
    SYSTEM_PACKAGES.append("tomli")


# Not using __pkgname__ because we want to be backwards compatible
TOML_SECTION_NAME = "pip-licenses"



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: Copyright (c) 2018 raimon
# SPDX-FileCopyrightText: Copyright (c) 2025 stefan6419846

if __name__ == "__main__":  # pragma: no cover
    from piplicenses.cli import main

    main()



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: Copyright (c) 2018 raimon
# SPDX-FileCopyrightText: Copyright (c) 2025 stefan6419846

from __future__ import annotations

import argparse
import codecs
import sys
from enum import Enum, auto
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, cast

from piplicenses_lib import FromArg, NoValueEnum

from piplicenses import __summary__, __version__
from piplicenses.constants import DEFAULT_OUTPUT_FIELDS, SUMMARY_OUTPUT_FIELDS, TOML_SECTION_NAME
from piplicenses.output import create_licenses_table, create_summary_table

if sys.version_info >= (3, 11):  # pragma: no cover
    import tomllib
else:  # pragma: no cover
    import tomli as tomllib


if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Sequence

open = open  # To allow monkey-patching.


class CustomNamespace(argparse.Namespace):
    from_: FromArg
    order: OrderArg
    format_: FormatArg
    summary: bool
    output_file: str
    ignore_packages: list[str]
    packages: list[str]
    with_system: bool
    with_authors: bool
    with_urls: bool
    with_description: bool
    with_license_file: bool
    with_license_files: bool
    no_license_path: bool
    with_notice_file: bool
    with_notice_files: bool
    with_other_files: bool
    filter_strings: bool
    filter_code_page: str
    partial_match: bool
    fail_on: str | None
    allow_only: str | None
    collect_all_failures: bool


def get_output_fields(args: CustomNamespace) -> list[str]:
    if args.summary:
        return list(SUMMARY_OUTPUT_FIELDS)

    output_fields = list(DEFAULT_OUTPUT_FIELDS)

    if args.from_ == FromArg.ALL:
        output_fields.append("License-Metadata")
        output_fields.append("License-Classifier")
    else:
        output_fields.append("License")

    if args.with_authors:
        output_fields.append("Author")

    if args.with_maintainers:
        output_fields.append("Maintainer")

    if args.with_urls:
        output_fields.append("URL")

    if args.with_description:
        output_fields.append("Description")

    if args.no_version:
        output_fields.remove("Version")

    if args.with_license_files and args.format_ not in [FormatArg.JSON, FormatArg.PLAIN_VERTICAL]:
        args.with_license_files = False
        args.with_notice_files = False
        args.with_other_files = False

    if args.with_license_file or args.with_license_files:
        if not args.no_license_path:
            output_fields.append("LicenseFiles" if args.with_license_files else "LicenseFile")

        output_fields.append("LicenseTexts" if args.with_license_files else "LicenseText")

        if args.with_notice_file or args.with_notice_files:
            if not args.no_license_path:
                output_fields.append("NoticeFiles" if args.with_notice_files else "NoticeFile")
            output_fields.append("NoticeTexts" if args.with_notice_files else "NoticeText")
        if args.with_other_files:
            if not args.no_license_path:
                output_fields.append("OtherFiles")
            output_fields.append("OtherTexts")

    return output_fields


def get_sortby(args: CustomNamespace) -> str:
    if args.summary and args.order == OrderArg.COUNT:
        return "Count"
    elif args.summary or args.order == OrderArg.LICENSE:
        return "License"
    elif args.order == OrderArg.NAME:
        return "Name"
    elif args.order == OrderArg.AUTHOR and args.with_authors:
        return "Author"
    elif args.order == OrderArg.MAINTAINER and args.with_maintainers:
        return "Maintainer"
    elif args.order == OrderArg.URL and args.with_urls:
        return "URL"

    return "Name"


def create_output_string(args: CustomNamespace) -> str:
    output_fields = get_output_fields(args)

    if args.summary:
        table = create_summary_table(args)
    else:
        table = create_licenses_table(args, output_fields)

    sortby = get_sortby(args)

    if args.format_ == FormatArg.HTML:
        html = table.get_html_string(fields=output_fields, sortby=sortby)
        return html.encode("ascii", errors="xmlcharrefreplace").decode("ascii")
    else:
        return table.get_string(fields=output_fields, sortby=sortby)


def create_warn_string(args: CustomNamespace) -> str:
    warn_messages = []
    warn = partial(output_colored, "33")

    if args.with_license_file and not args.format_ == FormatArg.JSON:
        message = warn("Due to the length of these fields, this option is best paired with --format=json.")
        warn_messages.append(message)

    if args.with_license_files and args.format_ not in [FormatArg.JSON, FormatArg.PLAIN_VERTICAL]:
        message = warn("Ignoring request to output multiple files due to unsupported output format.")
        warn_messages.append(message)

    if args.summary and (args.with_authors or args.with_urls):
        message = warn(
            "When using this option, only --order=count or --order=license has an effect for the --order "
            "option. And using --with-authors and --with-urls will be ignored."
        )
        warn_messages.append(message)

    return "\n".join(warn_messages)


class CustomHelpFormatter(argparse.HelpFormatter):  # pragma: no cover
    def __init__(
        self,
        prog: str,
        indent_increment: int = 2,
        max_help_position: int = 24,
        width: int | None = None,
    ) -> None:
        max_help_position = 30
        super().__init__(
            prog,
            indent_increment=indent_increment,
            max_help_position=max_help_position,
            width=width,
        )

    def _format_action(self, action: argparse.Action) -> str:
        flag_indent_argument: bool = False
        text = self._expand_help(action)
        separator_pos = text[:3].find("|")
        if separator_pos != -1 and "I" in text[:separator_pos]:
            self._indent()
            flag_indent_argument = True
        help_str = super()._format_action(action)
        if flag_indent_argument:
            self._dedent()
        return help_str

    def _expand_help(self, action: argparse.Action) -> str:
        if isinstance(action.default, Enum):
            default_value = enum_key_to_value(action.default)
            return cast(str, self._get_help_string(action)) % {"default": default_value}
        return super()._expand_help(action)

    def _split_lines(self, text: str, width: int) -> list[str]:
        separator_pos = text[:3].find("|")
        if separator_pos != -1:
            flag_splitlines: bool = "R" in text[:separator_pos]
            text = text[separator_pos + 1:]  # fmt: skip
            if flag_splitlines:
                return text.splitlines()
        return super()._split_lines(text, width)


class CompatibleArgumentParser(argparse.ArgumentParser):
    def parse_args(  # type: ignore[override]
        self,
        args: None | Sequence[str] = None,
        namespace: None | CustomNamespace = None,
    ) -> CustomNamespace:
        args_ = cast(CustomNamespace, super().parse_args(args, namespace))
        self._verify_args(args_)
        return args_

    def _verify_args(self, args: CustomNamespace) -> None:
        if args.with_license_file is False and args.with_license_files is False:
            if args.no_license_path is True or args.with_notice_file is True or args.with_notice_files is True or args.with_other_files is True:
                self.error(
                    "'--no-license-path' and '--with-notice-file[s]' as well as '--with-other-files' require the '--with-license-file[s]' option to be set"
                )
        if args.filter_strings is False and args.filter_code_page != "latin1":
            self.error("'--filter-code-page' requires the '--filter-strings' option to be set")
        try:
            codecs.lookup(args.filter_code_page)
        except LookupError:
            self.error(
                f"invalid code page {args.filter_code_page!r} given for '--filter-code-page, check "
                "https://docs.python.org/3/library/codecs.html#standard-encodings for valid code pages"
            )


class OrderArg(NoValueEnum):
    COUNT = C = auto()
    LICENSE = L = auto()
    NAME = N = auto()
    AUTHOR = A = auto()
    MAINTAINER = M = auto()
    URL = U = auto()


class FormatArg(NoValueEnum):
    PLAIN = P = auto()
    PLAIN_VERTICAL = auto()
    MARKDOWN = MD = M = auto()
    RST = REST = R = auto()
    CONFLUENCE = C = auto()
    HTML = H = auto()
    JSON = J = auto()
    JSON_LICENSE_FINDER = JLF = auto()
    CSV = auto()


def value_to_enum_key(value: str) -> str:
    return value.replace("-", "_").upper()


def enum_key_to_value(enum_key: Enum) -> str:
    return enum_key.name.replace("_", "-").lower()


def choices_from_enum(enum_cls: type[NoValueEnum]) -> list[str]:
    return [key.replace("_", "-").lower() for key in enum_cls.__members__.keys()]


def get_value_from_enum(enum_cls: type[NoValueEnum], value: str) -> NoValueEnum:
    return getattr(enum_cls, value_to_enum_key(value))


MAP_DEST_TO_ENUM = {
    "from_": FromArg,
    "order": OrderArg,
    "format_": FormatArg,
}


class SelectAction(argparse.Action):
    def __call__(  # type: ignore[override]
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str,
        option_string: str | None = None,
    ) -> None:
        enum_cls = MAP_DEST_TO_ENUM[self.dest]
        setattr(namespace, self.dest, get_value_from_enum(enum_cls, values))


def load_config_from_file(pyproject_path: str):
    path = Path(pyproject_path)
    if path.exists():
        with path.open(mode="rb") as f:
            return tomllib.load(f).get("tool", {}).get(TOML_SECTION_NAME, {})
    return {}


def create_parser(
    pyproject_path: str = "pyproject.toml",
) -> CompatibleArgumentParser:
    parser = CompatibleArgumentParser(description=__summary__, formatter_class=CustomHelpFormatter)

    config_from_file = load_config_from_file(pyproject_path)

    common_options = parser.add_argument_group("Common options")
    format_options = parser.add_argument_group("Format options")
    verify_options = parser.add_argument_group("Verify options")

    parser.add_argument("-v", "--version", action="version", version="%(prog)s " + __version__)

    common_options.add_argument(
        "--python",
        type=str,
        default=config_from_file.get("python", sys.executable),
        metavar="PYTHON_EXEC",
        help=(
            "R| path to python executable to search distributions from\n"
            "Package will be searched in the selected python's sys.path\n"
            "By default, will search packages for current env executable\n"
            "(default: sys.executable)"
        ),
    )

    common_options.add_argument(
        "--from",
        dest="from_",
        action=SelectAction,
        type=str,
        default=get_value_from_enum(FromArg, config_from_file.get("from", "mixed")),
        metavar="SOURCE",
        choices=choices_from_enum(FromArg),
        help='R|where to find license information\n"meta", "classifier, "mixed", "all"\n(default: %(default)s)',
    )
    common_options.add_argument(
        "-o",
        "--order",
        action=SelectAction,
        type=str,
        default=get_value_from_enum(OrderArg, config_from_file.get("order", "name")),
        metavar="COL",
        choices=choices_from_enum(OrderArg),
        help='R|order by column\n"name", "license", "author", "url"\n(default: %(default)s)',
    )
    common_options.add_argument(
        "-f",
        "--format",
        dest="format_",
        action=SelectAction,
        type=str,
        default=get_value_from_enum(FormatArg, config_from_file.get("format", "plain")),
        metavar="STYLE",
        choices=choices_from_enum(FormatArg),
        help=(
            "R|dump as set format style\n"
            '"plain", "plain-vertical" "markdown", "rst", \n'
            '"confluence", "html", "json", \n'
            '"json-license-finder",  "csv"\n'
            "(default: %(default)s)"
        ),
    )
    common_options.add_argument(
        "--summary",
        action="store_true",
        default=config_from_file.get("summary", False),
        help="dump summary of each license",
    )
    common_options.add_argument(
        "--output-file",
        action="store",
        default=config_from_file.get("output-file"),
        type=str,
        help="save license list to file",
    )
    common_options.add_argument(
        "-i",
        "--ignore-packages",
        action="store",
        type=str,
        nargs="+",
        metavar="PKG",
        default=config_from_file.get("ignore-packages", []),
        help="ignore package name in dumped list",
    )
    common_options.add_argument(
        "-p",
        "--packages",
        action="store",
        type=str,
        nargs="+",
        metavar="PKG",
        default=config_from_file.get("packages", []),
        help="only include selected packages in output",
    )
    format_options.add_argument(
        "-s",
        "--with-system",
        action="store_true",
        default=config_from_file.get("with-system", False),
        help="dump with system packages",
    )
    format_options.add_argument(
        "-a",
        "--with-authors",
        action="store_true",
        default=config_from_file.get("with-authors", False),
        help="dump with package authors",
    )
    format_options.add_argument(
        "--with-maintainers",
        action="store_true",
        default=config_from_file.get("with-maintainers", False),
        help="dump with package maintainers",
    )
    format_options.add_argument(
        "-u",
        "--with-urls",
        action="store_true",
        default=config_from_file.get("with-urls", False),
        help="dump with package urls",
    )
    format_options.add_argument(
        "-d",
        "--with-description",
        action="store_true",
        default=config_from_file.get("with-description", False),
        help="dump with short package description",
    )
    format_options.add_argument(
        "-nv",
        "--no-version",
        action="store_true",
        default=config_from_file.get("no-version", False),
        help="dump without package version",
    )
    format_options.add_argument(
        "-l",
        "--with-license-file",
        action="store_true",
        default=config_from_file.get("with-license-file", False),
        help="dump with location of license file and contents, most useful with JSON output",
    )
    format_options.add_argument(
        "--with-license-files",
        action="store_true",
        default=config_from_file.get("with-license-files", False),
        help="dump with location of license files and contents, most useful with JSON output",
    )
    format_options.add_argument(
        "--no-license-path",
        action="store_true",
        default=config_from_file.get("no-license-path", False),
        help="I|when specified together with option -l, suppress location of license file output",
    )
    format_options.add_argument(
        "--with-notice-file",
        action="store_true",
        default=config_from_file.get("with-notice-file", False),
        help="I|when specified together with option -l, dump with location of notice files and contents",
    )
    format_options.add_argument(
        "--with-notice-files",
        action="store_true",
        default=config_from_file.get("with-notice-files", False),
        help="I|when specified together with option -l or --with-license-files, dump with location of notice files and contents",
    )
    format_options.add_argument(
        "--with-other-files",
        action="store_true",
        default=config_from_file.get("with-other-files", False),
        help="I|when specified together with option -l or --with-license-files, dump with location of other licensing-related files and contents",
    )
    format_options.add_argument(
        "--filter-strings",
        action="store_true",
        default=config_from_file.get("filter-strings", False),
        help="filter input according to code page",
    )
    format_options.add_argument(
        "--filter-code-page",
        action="store",
        type=str,
        default=config_from_file.get("filter-code-page", "latin1"),
        metavar="CODE",
        help="I|specify code page for filtering (default: %(default)s)",
    )

    verify_options.add_argument(
        "--fail-on",
        action="store",
        type=str,
        default=config_from_file.get("fail-on", None),
        help="fail (exit with code 1) on the first occurrence of the licenses of the semicolon-separated list",
    )
    verify_options.add_argument(
        "--allow-only",
        action="store",
        type=str,
        default=config_from_file.get("allow-only", None),
        help="fail (exit with code 1) on the first occurrence of the licenses not in the semicolon-separated list",
    )
    verify_options.add_argument(
        "--partial-match",
        action="store_true",
        default=config_from_file.get("partial-match", False),
        help="enables partial matching for --allow-only/--fail-on",
    )
    verify_options.add_argument(
        "--collect-all-failures",
        action="store_true",
        default=config_from_file.get("collect-all-failures", False),
        help="collect all license failures and report them after processing all packages",
    )

    return parser


def output_colored(code: str, text: str, is_bold: bool = False) -> str:
    """
    Create function to output with color sequence
    """
    if is_bold:
        code = f"1;{code}"

    return f"\033[{code}m{text}\033[0m"


def save_if_needs(output_file: None | str, output_string: str) -> None:
    """
    Save to path given by args
    """
    if output_file is None:
        return

    try:
        with open(output_file, "w", encoding="utf-8") as f:
            f.write(output_string)
            if not output_string.endswith("\n"):
                # Always end output files with a new line
                f.write("\n")

        sys.stdout.write("created path: " + output_file + "\n")
        sys.exit(0)
    except OSError:
        sys.stderr.write("check path: --output-file\n")
        sys.exit(1)


def main() -> None:  # pragma: no cover
    parser = create_parser()
    args = parser.parse_args()

    output_string = create_output_string(args)

    output_file = args.output_file
    save_if_needs(output_file, output_string)

    print(output_string)
    warn_string = create_warn_string(args)
    if warn_string:
        print(warn_string, file=sys.stderr)



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: Copyright (c) 2018 raimon
# SPDX-FileCopyrightText: Copyright (c) 2025 stefan6419846

from unittest import TestCase

from piplicenses.cli import CustomNamespace
from piplicenses.collector import (
    _should_include_files,
    case_insensitive_partial_match_set_diff,
    case_insensitive_partial_match_set_intersect,
    case_insensitive_set_diff,
    case_insensitive_set_intersect,
    get_packages,
    parse_licenses_list,
)
from tests import UNICODE_APPENDIX, CommandLineTestCase, PatchDistributionsTestCase


class ShouldIncludeFilesTestCase(CommandLineTestCase):
    KEYS = ["with_license_file", "with_license_files", "with_notice_file", "with_notice_files", "with_other_files"]

    def test_none_requested(self) -> None:
        args = CustomNamespace(**{key: False for key in self.KEYS})
        self.assertFalse(_should_include_files(args))

    def test_one_requested(self) -> None:
        args = CustomNamespace()

        def set_keys(_enabled: str) -> None:
            for _key in self.KEYS:
                setattr(args, _key, _key == _enabled)

        for name in self.KEYS:
            with self.subTest(name=name):
                set_keys(name)
                self.assertTrue(_should_include_files(args))

    def test_multiple_requested(self) -> None:
        args = CustomNamespace(**{key: key.endswith("s") for key in self.KEYS})
        self.assertTrue(_should_include_files(args))


class GetPackagesTestCase(PatchDistributionsTestCase):
    def test_without_filter(self) -> None:
        self._patch_distributions()
        args = self.parser.parse_args([])
        packages = list(get_packages(args))
        self.assertIn(UNICODE_APPENDIX, packages[-1].name)

    def test_with_default_filter(self) -> None:
        self._patch_distributions()
        args = self.parser.parse_args(["--filter-strings"])
        packages = list(get_packages(args))
        self.assertNotIn(UNICODE_APPENDIX, packages[-1].name)

    def test_with_default_filter_and_license_file(self) -> None:
        self._patch_distributions()
        args = self.parser.parse_args(["--filter-strings", "--with-license-file"])
        packages = list(get_packages(args))
        self.assertNotIn(UNICODE_APPENDIX, packages[-1].name)

    def test_with_specified_filter(self) -> None:
        self._patch_distributions()
        args = self.parser.parse_args(["--filter-strings", "--filter-code-page=ascii"])
        packages = list(get_packages(args))
        self.assertNotIn(UNICODE_APPENDIX, packages[-1].summary)


class ParseLicensesListTestCase(TestCase):
    def test_parse_licenses_list(self) -> None:
        licenses_str = " MIT License;;  MIT    ;  Apache-2.0;;;    "

        licenses = parse_licenses_list(licenses_str)

        self.assertListEqual(["MIT License", "MIT", "Apache-2.0"], licenses)


class CaseInsensitivePartialMatchSetDiffTestCase(TestCase):
    def test_case_insensitive_partial_match_set_diff(self) -> None:
        set_a = {"MIT License"}
        set_b = {"Mit", "BSD License"}
        set_c = {"mit license"}
        a_diff_b = case_insensitive_partial_match_set_diff(set_a, set_b)
        a_diff_c = case_insensitive_partial_match_set_diff(set_a, set_c)
        b_diff_c = case_insensitive_partial_match_set_diff(set_b, set_c)
        a_diff_empty = case_insensitive_partial_match_set_diff(set_a, set())

        self.assertSetEqual(set(), a_diff_b)
        self.assertSetEqual(set(), a_diff_c)
        self.assertIn("BSD License", b_diff_c)
        self.assertIn("MIT License", a_diff_empty)

    def test_no_overlap(self) -> None:
        """
        When no overlap, the result should be the same as set_a.
        """
        set_a = {"Python", "Java", "C++"}
        set_b = {"Ruby", "JavaScript"}
        result = case_insensitive_partial_match_set_diff(set_a, set_b)
        self.assertSetEqual(set_a, result)

    def test_full_overlap(self) -> None:
        """
        When all items overlap, the result should be an empty set.
        """
        set_a = {"Hello", "World"}
        set_b = {"hello", "world"}
        result = case_insensitive_partial_match_set_diff(set_a, set_b)
        self.assertSetEqual(set(), result)

    def test_partial_overlap(self) -> None:
        """
        Only 'Python' should remain as it has no overlap with set_b.
        """
        set_a = {"HelloWorld", "Python", "JavaScript"}
        set_b = {"hello", "script"}
        result = case_insensitive_partial_match_set_diff(set_a, set_b)
        self.assertSetEqual({"Python"}, result)

    def test_case_insensitive(self) -> None:
        """
        The function should handle case-insensitive matches correctly.
        """
        set_a = {"HELLO", "world"}
        set_b = {"hello"}
        result = case_insensitive_partial_match_set_diff(set_a, set_b)
        self.assertSetEqual({"world"}, result)

    def test_empty_sets(self) -> None:
        """
        When both sets are empty, the result should also be empty.
        """
        set_a: set[str] = set()
        set_b: set[str] = set()
        result = case_insensitive_partial_match_set_diff(set_a, set_b)
        self.assertSetEqual(set(), result)

    def test_set_b_is_empty(self) -> None:
        """
        If set_b is empty, result should be the same as set_a.
        """
        set_a = {"Python", "Java"}
        set_b: set[str] = set()
        result = case_insensitive_partial_match_set_diff(set_a, set_b)
        self.assertSetEqual(set_a, result)

    def test_set_a_is_empty(self) -> None:
        """
        If set_a is empty, result should be empty regardless of set_b.
        """
        set_a: set[str] = set()
        set_b = {"Ruby"}
        result = case_insensitive_partial_match_set_diff(set_a, set_b)
        self.assertSetEqual(set(), result)

    def test_set_a_is_not_empty(self) -> None:
        """
        If set_b contains two similar substrings, there should be no double deletion.
        """
        set_a = {"Python 3.11", "Python 3.12", "Javascript"}
        set_b = {"Python", "Python 3"}
        result = case_insensitive_partial_match_set_diff(set_a, set_b)
        self.assertSetEqual({"Javascript"}, result)


class CaseInsensitiveSetDiffTestCase(TestCase):
    def test_case_insensitive_set_diff(self) -> None:
        set_a = {"MIT License"}
        set_b = {"Mit License", "BSD License"}
        set_c = {"mit license"}
        a_diff_b = case_insensitive_set_diff(set_a, set_b)
        a_diff_c = case_insensitive_set_diff(set_a, set_c)
        b_diff_c = case_insensitive_set_diff(set_b, set_c)
        a_diff_empty = case_insensitive_set_diff(set_a, set())

        self.assertSetEqual(set(), a_diff_b)
        self.assertSetEqual(set(), a_diff_c)
        self.assertIn("BSD License", b_diff_c)
        self.assertIn("MIT License", a_diff_empty)


class CaseInsensitiveSetIntersectTestCase(TestCase):
    def test_case_insensitive_set_intersect(self) -> None:
        set_a = {"Revised BSD"}
        set_b = {"Apache License", "revised BSD"}
        set_c = {"revised bsd"}
        a_intersect_b = case_insensitive_set_intersect(set_a, set_b)
        a_intersect_c = case_insensitive_set_intersect(set_a, set_c)
        b_intersect_c = case_insensitive_set_intersect(set_b, set_c)
        a_intersect_empty = case_insensitive_set_intersect(set_a, set())

        self.assertSetEqual(set_a, a_intersect_b)
        self.assertSetEqual(set_a, a_intersect_c)
        self.assertSetEqual({"revised BSD"}, b_intersect_c)
        self.assertSetEqual(set(), a_intersect_empty)


class CaseInsensitivePartialMatchSetIntersectTestCase(TestCase):
    def test_case_insensitive_partial_match_set_intersect(self) -> None:
        set_a = {"Revised BSD"}
        set_b = {"Apache License", "revised BSD"}
        set_c = {"bsd"}
        a_intersect_b = case_insensitive_partial_match_set_intersect(set_a, set_b)
        a_intersect_c = case_insensitive_partial_match_set_intersect(set_a, set_c)
        b_intersect_c = case_insensitive_partial_match_set_intersect(set_b, set_c)
        a_intersect_empty = case_insensitive_partial_match_set_intersect(set_a, set())

        self.assertSetEqual(set_a, a_intersect_b)
        self.assertSetEqual(set_a, a_intersect_c)
        self.assertSetEqual({"revised BSD"}, b_intersect_c)
        self.assertSetEqual(set(), a_intersect_empty)



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: Copyright (c) 2018 raimon
# SPDX-FileCopyrightText: Copyright (c) 2025 stefan6419846

from docutils.frontend import get_default_settings
from docutils.parsers.rst import Parser
from docutils.utils import new_document
from prettytable import HRuleStyle

from piplicenses.output import create_licenses_table, factory_styled_table_with_args
from tests import PatchDistributionsTestCase


class CreateLicensesTableTestCase(PatchDistributionsTestCase):
    @classmethod
    def check_rst(cls, text: str) -> None:
        parser = Parser()
        settings = get_default_settings(parser)
        settings.halt_level = 3
        document = new_document("<rst-doc>", settings=settings)
        parser.parse(text, document)

    def test_format_markdown(self) -> None:
        format_markdown_args = ["--format=markdown"]
        args = self.parser.parse_args(format_markdown_args)
        table = create_licenses_table(args)

        self.assertIn("l", table.align.values())
        self.assertTrue(table.border)
        self.assertTrue(table.header)
        self.assertEqual("|", table.junction_char)
        self.assertEqual(HRuleStyle.HEADER, table.hrules)

    def test_format_rst_without_filter(self) -> None:
        self._patch_distributions()
        format_rst_args = ["--format=rst"]
        args = self.parser.parse_args(format_rst_args)
        table = create_licenses_table(args)

        self.assertIn("l", table.align.values())
        self.assertTrue(table.border)
        self.assertTrue(table.header)
        self.assertEqual("+", table.junction_char)
        self.assertEqual(HRuleStyle.ALL, table.hrules)

    def test_format_rst_default_filter(self) -> None:
        self._patch_distributions()
        format_rst_args = ["--format=rst", "--filter-strings"]
        args = self.parser.parse_args(format_rst_args)
        table = create_licenses_table(args)

        self.assertIn("l", table.align.values())
        self.assertTrue(table.border)
        self.assertTrue(table.header)
        self.assertEqual("+", table.junction_char)
        self.assertEqual(HRuleStyle.ALL, table.hrules)
        self.check_rst(str(table))

    def test_format_confluence(self) -> None:
        format_confluence_args = ["--format=confluence"]
        args = self.parser.parse_args(format_confluence_args)
        table = create_licenses_table(args)

        self.assertIn("l", table.align.values())
        self.assertTrue(table.border)
        self.assertTrue(table.header)
        self.assertEqual("|", table.junction_char)
        self.assertEqual(HRuleStyle.NONE, table.hrules)


class FactoryStyledTableWithArgsTestCase(PatchDistributionsTestCase):
    def test_format_plain(self) -> None:
        format_plain_args = ["--format=plain"]
        args = self.parser.parse_args(format_plain_args)
        table = factory_styled_table_with_args(args)

        self.assertIn("l", table.align.values())
        self.assertFalse(table.border)
        self.assertTrue(table.header)
        self.assertEqual("+", table.junction_char)
        self.assertEqual(HRuleStyle.FRAME, table.hrules)



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: Copyright (c) 2018 raimon
# SPDX-FileCopyrightText: Copyright (c) 2025 stefan6419846

from __future__ import annotations

import sys
from dataclasses import asdict
from typing import TYPE_CHECKING, cast

from piplicenses_lib import get_packages as _get_packages
from piplicenses_lib import normalize_package_name

from piplicenses.constants import SYSTEM_PACKAGES
from piplicenses.spdx import _parse_spdx

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Iterator

    from piplicenses_lib import PackageInfo

    from piplicenses.cli import CustomNamespace


def parse_licenses_list(licenses_str: str | None) -> list[str]:
    if licenses_str is None:
        return []

    licenses = licenses_str.split(";")

    # Strip items
    licenses = list(map(str.strip, licenses))

    # Remove empty string items
    licenses = list(filter(None, licenses))

    return licenses


def _should_include_files(args: CustomNamespace) -> bool:
    return args.with_license_file or args.with_license_files or args.with_notice_file or args.with_notice_files or args.with_other_files


def get_packages(
    args: CustomNamespace,
) -> Iterator[PackageInfo]:
    ignore_pkgs_as_normalize = list(map(normalize_package_name, args.ignore_packages))
    pkgs_as_normalize = list(map(normalize_package_name, args.packages))

    fail_on_licenses = set(parse_licenses_list(args.fail_on))
    allow_only_licenses = set(parse_licenses_list(args.allow_only))

    include_files = _should_include_files(args)
    failures = []
    for pkg_info in _get_packages(
        from_source=args.from_,
        python_path=args.python,
        include_files=include_files,
        normalize_names=False,
    ):
        pkg_name = normalize_package_name(pkg_info.name)
        pkg_name_and_version = pkg_name + ":" + pkg_info.version

        if pkg_name.lower() in ignore_pkgs_as_normalize or pkg_name_and_version.lower() in ignore_pkgs_as_normalize:
            continue

        if pkgs_as_normalize and pkg_name.lower() not in pkgs_as_normalize:
            continue

        if not args.with_system and pkg_name in SYSTEM_PACKAGES:
            continue

        if args.filter_strings:

            def filter_string(item: str) -> str:
                return item.encode(args.filter_code_page, errors="ignore").decode(args.filter_code_page)

            for key, value in asdict(pkg_info).items():
                if key == "distribution":
                    continue

                def _handle(_value):
                    if isinstance(_value, list):
                        return list(map(_handle, _value))
                    if isinstance(_value, set):
                        return set(map(_handle, _value))
                    if isinstance(_value, tuple):
                        return tuple(map(_handle, _value))
                    return filter_string(cast(str, _value))

                setattr(pkg_info, key, _handle(value))

        parsed_license_names: set[str] = set()
        for license_expr in pkg_info.license_names:
            parsed_license_names |= _parse_spdx(license_expr)

        fail_this_pkg = False
        fail_message = ""
        if fail_on_licenses:
            if not args.partial_match:
                failed_licenses = case_insensitive_set_intersect(parsed_license_names, fail_on_licenses)
            else:
                failed_licenses = case_insensitive_partial_match_set_intersect(parsed_license_names, fail_on_licenses)
            if failed_licenses:
                fail_this_pkg = True
                fail_message = "fail-on license {} was found for package {}:{}\n".format(
                    "; ".join(sorted(failed_licenses)),
                    pkg_info.name,
                    pkg_info.version,
                )
        if allow_only_licenses:
            if not args.partial_match:
                uncommon_licenses = case_insensitive_set_diff(parsed_license_names, allow_only_licenses)
            else:
                uncommon_licenses = case_insensitive_partial_match_set_diff(parsed_license_names, allow_only_licenses)
            if len(uncommon_licenses) == len(parsed_license_names):
                fail_this_pkg = True
                fail_message = "license {} not in allow-only licenses was found for package {}:{}\n".format(
                    "; ".join(sorted(uncommon_licenses)),
                    pkg_info.name,
                    pkg_info.version,
                )
        if fail_this_pkg:
            if args.collect_all_failures:
                failures.append(fail_message)
            else:
                sys.stderr.write(fail_message)
                sys.exit(1)

        yield pkg_info

    if args.collect_all_failures and failures:
        for msg in failures:
            sys.stderr.write(msg)
        sys.exit(1)


def case_insensitive_set_intersect(set_a, set_b):
    """Same as set.intersection() but case-insensitive"""
    common_items = set()
    set_b_lower = {item.lower() for item in set_b}
    for elem in set_a:
        if elem.lower() in set_b_lower:
            common_items.add(elem)
    return common_items


def case_insensitive_partial_match_set_intersect(set_a, set_b):
    common_items = set()
    for item_a in set_a:
        for item_b in set_b:
            if item_b.lower() in item_a.lower():
                common_items.add(item_a)
    return common_items


def case_insensitive_partial_match_set_diff(set_a, set_b):
    uncommon_items = set_a.copy()
    for item_a in set_a:
        for item_b in set_b:
            if item_b.lower() in item_a.lower():
                uncommon_items.discard(item_a)
    return uncommon_items


def case_insensitive_set_diff(set_a, set_b):
    """Same as set.difference() but case-insensitive"""
    uncommon_items = set()
    set_b_lower = {item.lower() for item in set_b}
    for elem in set_a:
        if elem.lower() not in set_b_lower:
            uncommon_items.add(elem)
    return uncommon_items



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: Copyright (c) 2018 raimon
# SPDX-FileCopyrightText: Copyright (c) 2025 stefan6419846

import warnings
from unittest import TestCase

from piplicenses.errors import PipLicensesWarning
from piplicenses.spdx import _get_spdx_parser

try:
    import license_expression
except ImportError:
    license_expression = None  # type: ignore[assignment]


class SpdxParserTestCase(TestCase):
    def test_has_license_expression_package(self):
        if license_expression is None:
            raise self.skipTest("Requires license-expression package.")

        for test_id, expression, expected, should_warn in [
            (
                "SPDX WITH",
                "GPL-2.0-or-later WITH Bison-exception-2.2",
                {"GPL-2.0-or-later WITH Bison-exception-2.2"},
                True,
            ),
            ("SPDX AND", "Apache-2.0 AND BSD-3-Clause", {"Apache-2.0 AND BSD-3-Clause"}, True),
            ("SPDX OR", "Apache-2.0 OR BSD-3-Clause", {"Apache-2.0", "BSD-3-Clause"}, False),
            ("Invalid SPDX", "Hello World", {"Hello World"}, False),
            ("Empty", "", {""}, False),
        ]:
            with self.subTest(test_id=test_id, expression=expression):
                if should_warn:
                    with warnings.catch_warnings(record=True) as caught_warnings:
                        warnings.simplefilter("always")
                        self.assertEqual(expected, _get_spdx_parser()(expression))
                        self.assertEqual(1, len(caught_warnings), caught_warnings)
                        self.assertEqual(PipLicensesWarning, caught_warnings[0].category)
                else:
                    self.assertEqual(expected, _get_spdx_parser()(expression))

    def test_does_not_have_license_expression_package(self):
        if license_expression is not None:
            raise self.skipTest("Does not work with license-expression package.")

        for (
            test_id,
            expression,
        ) in [
            ("SPDX WITH", "GPL-2.0-or-later WITH Bison-exception-2.2"),
            ("SPDX AND", "Apache-2.0 AND BSD-3-Clause"),
            ("SPDX OR", "Apache-2.0 OR BSD-3-Clause"),
            ("Invalid SPDX", "Hello World"),
            ("Empty", ""),
        ]:
            with self.subTest(test_id=test_id, expression=expression):
                self.assertEqual({expression}, _get_spdx_parser()(expression))



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: Copyright (c) 2018 raimon
# SPDX-FileCopyrightText: Copyright (c) 2025 stefan6419846

import copy
import json
import sys
import venv
from importlib.metadata import Distribution, PathDistribution
from tempfile import TemporaryDirectory
from types import SimpleNamespace
from unittest import mock

from piplicenses_lib import PackageInfo, normalize_package_name
from prettytable import HRuleStyle, PrettyTable

from piplicenses import __pkgname__
from piplicenses.cli import create_output_string, create_parser, create_warn_string, get_output_fields, get_sortby
from piplicenses.collector import get_packages
from piplicenses.constants import DEFAULT_OUTPUT_FIELDS, SYSTEM_PACKAGES
from piplicenses.output import create_licenses_table
from tests import CaptureOutput, CommandLineTestCase

try:
    import license_expression
except ImportError:
    license_expression = None  # type: ignore[assignment]


CRYPTOGRAPHY_VERSION = Distribution.from_name("cryptography").version


class IntegrationTestCase(CommandLineTestCase):
    @classmethod
    def _create_pkg_name_columns(cls, table: PrettyTable) -> list[str]:
        index = DEFAULT_OUTPUT_FIELDS.index("Name")

        rows = copy.deepcopy(table.rows)
        pkg_name_columns = []
        for row in rows:
            pkg_name_columns.append(row[index])

        return pkg_name_columns

    @classmethod
    def _create_license_columns(cls, table: PrettyTable, output_fields: list[str]) -> list[str]:
        index = output_fields.index("License")

        rows = copy.deepcopy(table.rows)
        pkg_name_columns = []
        for row in rows:
            pkg_name_columns.append(row[index])

        return pkg_name_columns

    def test_with_empty_args(self) -> None:
        empty_args: list[str] = []
        args = self.parser.parse_args(empty_args)
        table = create_licenses_table(args)

        self.assertIn("l", table.align.values())
        self.assertFalse(table.border)
        self.assertTrue(table.header)
        self.assertEqual("+", table.junction_char)
        self.assertEqual(HRuleStyle.FRAME, table.hrules)

        output_fields = get_output_fields(args)
        self.assertEqual(output_fields, list(DEFAULT_OUTPUT_FIELDS) + ["License"])
        self.assertNotIn("Author", output_fields)
        self.assertNotIn("URL", output_fields)

        pkg_name_columns = self._create_pkg_name_columns(table)
        for sys_pkg in SYSTEM_PACKAGES:
            self.assertNotIn(sys_pkg, pkg_name_columns)

        sortby = get_sortby(args)
        self.assertEqual("Name", sortby)

        output_string = create_output_string(args)
        self.assertNotIn("<table>", output_string)

    def test_from_meta(self) -> None:
        from_args = ["--from=meta"]
        args = self.parser.parse_args(from_args)

        output_fields = get_output_fields(args)
        self.assertIn("License", output_fields)

        table = create_licenses_table(args, output_fields)
        license_columns = self._create_license_columns(table, output_fields)
        license_notation_as_meta = "MIT"
        self.assertIn(license_notation_as_meta, license_columns)

    def test_with_spdx_license_in_meta(self) -> None:
        from_args = ["--from=meta"]
        args = self.parser.parse_args(from_args)
        output_fields = get_output_fields(args)
        table = create_licenses_table(args, output_fields)

        self.assertIn("License", output_fields)

        license_columns = self._create_license_columns(table, output_fields)
        license_notation_as_classifier = "Apache-2.0 OR BSD-3-Clause"
        self.assertIn(license_notation_as_classifier, license_columns)

    def test_from_classifier(self) -> None:
        from_args = ["--from=classifier"]
        args = self.parser.parse_args(from_args)
        output_fields = get_output_fields(args)
        table = create_licenses_table(args, output_fields)

        self.assertIn("License", output_fields)

        license_columns = self._create_license_columns(table, output_fields)
        license_notation_as_classifier = "MIT License"
        self.assertIn(license_notation_as_classifier, license_columns)

    def test_from_mixed(self) -> None:
        from_args = ["--from=mixed"]
        args = self.parser.parse_args(from_args)
        output_fields = get_output_fields(args)
        table = create_licenses_table(args, output_fields)

        self.assertIn("License", output_fields)

        license_columns = self._create_license_columns(table, output_fields)
        # Depending on the condition "MIT" or "BSD" etc.
        license_notation_as_classifier = "MIT License"
        self.assertIn(license_notation_as_classifier, license_columns)

    def test_from_all(self) -> None:
        from_args = ["--from=all"]
        args = self.parser.parse_args(from_args)
        output_fields = get_output_fields(args)
        table = create_licenses_table(args, output_fields)

        self.assertIn("License-Metadata", output_fields)
        self.assertIn("License-Classifier", output_fields)

        index_license_meta = output_fields.index("License-Metadata")
        license_meta = []
        for row in table.rows:
            license_meta.append(row[index_license_meta])

        index_license_classifier = output_fields.index("License-Classifier")
        license_classifier = []
        for row in table.rows:
            license_classifier.append(row[index_license_classifier])

        for license_name in ("BSD-3-Clause", "MIT", "Apache 2.0"):
            self.assertIn(license_name, license_meta)
        for license_name in (
            "BSD License",
            "MIT License",
            "Apache Software License",
        ):
            self.assertIn(license_name, license_classifier)

    def test_with_system(self) -> None:
        with_system_args = ["--with-system"]
        args = self.parser.parse_args(with_system_args)
        table = create_licenses_table(args)

        pkg_name_columns = self._create_pkg_name_columns(table)
        pkg_names = list(map(normalize_package_name, pkg_name_columns))
        external_sys_pkgs = list(SYSTEM_PACKAGES)
        external_sys_pkgs.remove(__pkgname__)
        for sys_pkg in external_sys_pkgs:
            self.assertIn(sys_pkg, pkg_names)

    def test_with_authors(self) -> None:
        with_authors_args = ["--with-authors"]
        args = self.parser.parse_args(with_authors_args)

        output_fields = get_output_fields(args)
        self.assertNotEqual(output_fields, list(DEFAULT_OUTPUT_FIELDS))
        self.assertIn("Author", output_fields)

        output_string = create_output_string(args)
        self.assertIn("Author", output_string)

    def test_with_maintainers(self) -> None:
        with_maintainers_args = ["--with-maintainers"]
        args = self.parser.parse_args(with_maintainers_args)

        output_fields = get_output_fields(args)
        self.assertNotEqual(output_fields, list(DEFAULT_OUTPUT_FIELDS))
        self.assertIn("Maintainer", output_fields)

        output_string = create_output_string(args)
        self.assertIn("Maintainer", output_string)

    def test_with_urls(self) -> None:
        with_urls_args = ["--with-urls"]
        args = self.parser.parse_args(with_urls_args)

        output_fields = get_output_fields(args)
        self.assertNotEqual(output_fields, list(DEFAULT_OUTPUT_FIELDS))
        self.assertIn("URL", output_fields)

        output_string = create_output_string(args)
        self.assertIn("URL", output_string)

    def test_with_description(self) -> None:
        with_description_args = ["--with-description"]
        args = self.parser.parse_args(with_description_args)

        output_fields = get_output_fields(args)
        self.assertNotEqual(output_fields, list(DEFAULT_OUTPUT_FIELDS))
        self.assertIn("Description", output_fields)

        output_string = create_output_string(args)
        self.assertIn("Description", output_string)

    def test_without_version(self) -> None:
        without_version_args = ["--no-version"]
        args = self.parser.parse_args(without_version_args)

        output_fields = get_output_fields(args)
        self.assertNotEqual(output_fields, list(DEFAULT_OUTPUT_FIELDS))
        self.assertNotIn("Version", output_fields)

        output_string = create_output_string(args)
        self.assertNotIn("Version", output_string)

    def test_with_license_file(self) -> None:
        with_license_file_args = ["--with-license-file"]
        args = self.parser.parse_args(with_license_file_args)

        output_fields = get_output_fields(args)
        self.assertNotEqual(output_fields, list(DEFAULT_OUTPUT_FIELDS))
        self.assertIn("LicenseFile", output_fields)
        self.assertIn("LicenseText", output_fields)
        self.assertNotIn("NoticeFile", output_fields)
        self.assertNotIn("NoticeText", output_fields)

        output_string = create_output_string(args)
        self.assertIn("LicenseFile", output_string)
        self.assertIn("LicenseText", output_string)
        self.assertNotIn("NoticeFile", output_string)
        self.assertNotIn("NoticeText", output_string)

    def test_with_notice_file(self) -> None:
        with_license_file_args = ["--with-license-file", "--with-notice-file"]
        args = self.parser.parse_args(with_license_file_args)

        output_fields = get_output_fields(args)
        self.assertNotEqual(output_fields, list(DEFAULT_OUTPUT_FIELDS))
        self.assertIn("LicenseFile", output_fields)
        self.assertIn("LicenseText", output_fields)
        self.assertIn("NoticeFile", output_fields)
        self.assertIn("NoticeText", output_fields)

        output_string = create_output_string(args)
        self.assertIn("LicenseFile", output_string)
        self.assertIn("LicenseText", output_string)
        self.assertIn("NoticeFile", output_string)
        self.assertIn("NoticeText", output_string)

    def test_with_license_file_no_path(self) -> None:
        with_license_file_args = [
            "--with-license-file",
            "--with-notice-file",
            "--no-license-path",
        ]
        args = self.parser.parse_args(with_license_file_args)

        output_fields = get_output_fields(args)
        self.assertNotEqual(output_fields, list(DEFAULT_OUTPUT_FIELDS))
        self.assertNotIn("LicenseFile", output_fields)
        self.assertIn("LicenseText", output_fields)
        self.assertNotIn("NoticeFile", output_fields)
        self.assertIn("NoticeText", output_fields)

        output_string = create_output_string(args)
        self.assertNotIn("LicenseFile", output_string)
        self.assertIn("LicenseText", output_string)
        self.assertNotIn("NoticeFile", output_string)
        self.assertIn("NoticeText", output_string)

    def test_with_license_file_warning(self) -> None:
        with_license_file_args = ["--with-license-file", "--format=markdown"]
        args = self.parser.parse_args(with_license_file_args)

        warn_string = create_warn_string(args)
        self.assertIn("best paired with --format=json", warn_string)

    def test_ignore_packages(self) -> None:
        ignore_pkg_name = "prettytable"
        ignore_packages_args = [
            "--ignore-package=" + ignore_pkg_name,
            "--with-system",
        ]
        args = self.parser.parse_args(ignore_packages_args)
        table = create_licenses_table(args)

        pkg_name_columns = self._create_pkg_name_columns(table)
        self.assertNotIn(ignore_pkg_name, pkg_name_columns)

    def test_ignore_normalized_packages(self) -> None:
        ignore_pkg_name = "pip-licenses"
        ignore_packages_args = [
            "--ignore-package=pip_licenses",
            "--with-system",
        ]
        args = self.parser.parse_args(ignore_packages_args)
        table = create_licenses_table(args)

        pkg_name_columns = self._create_pkg_name_columns(table)
        self.assertNotIn(ignore_pkg_name, pkg_name_columns)

    def test_ignore_packages_and_version(self) -> None:
        # Fictitious version that does not exist
        ignore_pkg_name = "prettytable"
        ignore_pkg_spec = ignore_pkg_name + ":1.99.99"
        ignore_packages_args = [
            "--ignore-package=" + ignore_pkg_spec,
            "--with-system",
        ]
        args = self.parser.parse_args(ignore_packages_args)
        table = create_licenses_table(args)

        pkg_name_columns = self._create_pkg_name_columns(table)
        # It is expected that prettytable will include
        self.assertIn(ignore_pkg_name, pkg_name_columns)

    def test_with_packages(self) -> None:
        pkg_name = "pycodestyle"
        only_packages_args = ["--packages=" + pkg_name]
        args = self.parser.parse_args(only_packages_args)
        table = create_licenses_table(args)

        pkg_name_columns = self._create_pkg_name_columns(table)
        self.assertListEqual([pkg_name], pkg_name_columns)

    def test_with_normalized_packages(self) -> None:
        pkg_name = "typing_extensions"
        only_packages_args = [
            "--package=typing-extensions",
            "--with-system",
        ]
        args = self.parser.parse_args(only_packages_args)
        table = create_licenses_table(args)

        pkg_name_columns = self._create_pkg_name_columns(table)
        self.assertListEqual([pkg_name], pkg_name_columns)

    def test_with_packages_with_system(self) -> None:
        pkg_name = "prettytable"
        only_packages_args = ["--packages=" + pkg_name, "--with-system"]
        args = self.parser.parse_args(only_packages_args)
        table = create_licenses_table(args)

        pkg_name_columns = self._create_pkg_name_columns(table)
        self.assertListEqual([pkg_name], pkg_name_columns)

    def test_fail_on_partial_match(self) -> None:
        licenses = ("MIT",)
        allow_only_args = [
            "--partial-match",
            f"--fail-on={';'.join(licenses)}",
        ]

        with mock.patch("sys.exit"), CaptureOutput() as captured:
            args = create_parser().parse_args(allow_only_args)
            create_licenses_table(args)

        self.assertEqual("", captured.stdout)
        self.assertIn("fail-on license MIT License was found for package", captured.stderr)

    def test_allow_only(self) -> None:
        licenses = (
            "Bsd License",
            "Apache Software License",
            "Mozilla Public License 2.0 (MPL 2.0)",
            "Python Software Foundation License",
            "Public Domain",
            "GNU General Public License (GPL)",
            "GNU Library or Lesser General Public License (LGPL)",
        )
        allow_only_args = [f"--allow-only={';'.join(licenses)}"]

        with mock.patch("sys.exit"), CaptureOutput() as captured:
            args = create_parser().parse_args(allow_only_args)
            create_licenses_table(args)

        self.assertEqual("", captured.stdout)
        self.assertIn("license MIT License not in allow-only licenses was found for package", captured.stderr)

    def test_allow_only_collect_all_failures(self) -> None:
        licenses = (
            "Apache Software License",
            "Mozilla Public License 2.0 (MPL 2.0)",
            "Python Software Foundation License",
            "Public Domain",
            "GNU General Public License (GPL)",
            "GNU Library or Lesser General Public License (LGPL)",
        )
        allow_only_args = [
            f"--allow-only={';'.join(licenses)}",
            "--collect-all-failures",
        ]

        with mock.patch("sys.exit"), CaptureOutput() as captured:
            args = create_parser().parse_args(allow_only_args)
            create_licenses_table(args)

        self.assertIn("license MIT License not in allow-only licenses was found for package", captured.stderr)
        self.assertIn("license BSD License not in allow-only licenses was found for package", captured.stderr)

    def test_allow_only_partial(self) -> None:
        licenses = (
            "Bsd",
            "Apache",
            "Mozilla Public License 2.0 (MPL 2.0)",
            "Python Software Foundation License",
            "Public Domain",
            "GNU General Public License (GPL)",
            "GNU Library or Lesser General Public License (LGPL)",
        )
        allow_only_args = [
            "--partial-match",
            f"--allow-only={';'.join(licenses)}",
        ]

        with mock.patch("sys.exit"), CaptureOutput() as captured:
            args = create_parser().parse_args(allow_only_args)
            create_licenses_table(args)

        self.assertEqual("", captured.stdout)
        self.assertIn("license MIT License not in allow-only licenses was found for package", captured.stderr)

    def test_different_python(self) -> None:
        class TempEnvBuild(venv.EnvBuilder):
            def post_setup(self, context: SimpleNamespace) -> None:
                self.context = context

        with TemporaryDirectory() as target_dir_path:
            venv_builder = TempEnvBuild(with_pip=True)
            venv_builder.create(str(target_dir_path))
            python_exec = venv_builder.context.env_exe
            python_arg = f"--python={python_exec}"
            args = create_parser().parse_args([python_arg, "-s", "-f=json"])
            pkgs = get_packages(args)
            package_names = sorted({p.name for p in pkgs})

        expected_packages = ["pip"]
        if sys.version_info < (3, 12, 0):
            expected_packages.append("setuptools")
        assert package_names == expected_packages

    def test_fail_on(self) -> None:
        licenses = ("MIT license",)
        allow_only_args = [f"--fail-on={';'.join(licenses)}"]

        with mock.patch("sys.exit"), CaptureOutput() as captured:
            args = create_parser().parse_args(allow_only_args)
            create_licenses_table(args)

        self.assertEqual("", captured.stdout)
        self.assertIn("fail-on license MIT License was found for package", captured.stderr)

    def test_fail_on_collect_all_failures(self) -> None:
        licenses = ("MIT License", "BSD License")
        fail_on_args = [
            f"--fail-on={';'.join(licenses)}",
            "--collect-all-failures",
        ]

        with mock.patch("sys.exit"), CaptureOutput() as captured:
            args = create_parser().parse_args(fail_on_args)
            create_licenses_table(args)

        self.assertIn("fail-on license MIT License was found for package", captured.stderr)
        self.assertIn("fail-on license BSD License was found for package", captured.stderr)

    def test_spdx_operator_or_succeeds_if_either_license_is_allowed(self) -> None:
        # cryptography has an "Apache-2.0 OR BSD-3-Clause" license SPDX expression
        for expression in ["Apache-2.0", "BSD-3-Clause"]:
            spdx_args_success = [
                f"--allow-only={expression}",
                "--packages=cryptography",
            ]

            with mock.patch("sys.exit"), CaptureOutput() as captured:
                args = create_parser().parse_args(spdx_args_success)
                create_licenses_table(args)

            if license_expression is not None:
                self.assertEqual("", captured.stderr)
            else:
                self.assertEqual(
                    f"license Apache-2.0 OR BSD-3-Clause not in allow-only licenses was found for package cryptography:{CRYPTOGRAPHY_VERSION}\n",
                    captured.stderr,
                )

    def test_spdx_operator_or_fails_if_either_license_is_not_allowed(self) -> None:
        # cryptography has an "Apache-2.0 OR BSD-3-Clause" license SPDX expression
        for expression in ["Apache-2.0", "BSD-3-Clause"]:
            spdx_args_failure = [
                f"--fail-on={expression}",
                "--packages=cryptography",
            ]
            with mock.patch("sys.exit"), CaptureOutput() as captured:
                args = create_parser().parse_args(spdx_args_failure)
                create_licenses_table(args)

            if license_expression is None:
                self.assertEqual("", captured.stderr)
            else:
                self.assertEqual(f"fail-on license {expression} was found for package cryptography:{CRYPTOGRAPHY_VERSION}\n", captured.stderr)

    def test_file_handling(self) -> None:
        package = PackageInfo(
            name="test-package",
            version="1.0",
            distribution=PathDistribution.from_name("pip-licenses-lib"),
            licenses=[
                ("/django/contrib/admin/static/admin/css/vendor/select2/LICENSE-SELECT2.md", "This is the SELECT2 license text."),
                ("/django-5.2.6.dist-info/licenses/LICENSE", "This is the primary license text."),
                ("/django-5.2.6.dist-info/licenses/LICENSE.python", "This is the license for upstream Python."),
            ],
            notices=[
                ("/django-5.2.6.dist-info/licenses/NOTICE1", "This is the first notice file."),
                ("/django-5.2.6.dist-info/licenses/NOTICE2", "This is the second notice file."),
            ],
            others=[
                ("/django-5.2.6.dist-info/licenses/AUTHORS", "This is the AUTHORS file."),
                ("/django-5.2.6.dist-info/licenses/TESTING", "This is a dummy file for testing."),
            ],
        )

        with mock.patch("piplicenses.collector._get_packages", return_value=[package]):
            args = create_parser().parse_args(["--format=plain-vertical"])
            result = create_output_string(args)
            self.assertEqual("test-package\n1.0\n\n\n", result)

            args = create_parser().parse_args(["--format=plain-vertical", "--with-license-file", "--with-notice-file"])
            result = create_output_string(args)
            self.assertEqual(
                """test-package
1.0

/django/contrib/admin/static/admin/css/vendor/select2/LICENSE-SELECT2.md
This is the SELECT2 license text.
/django-5.2.6.dist-info/licenses/NOTICE1
This is the first notice file.

""",
                result,
            )

            args = create_parser().parse_args(["--format=plain-vertical", "--with-license-files", "--with-notice-files", "--with-other-files"])
            result = create_output_string(args)
            self.assertEqual(
                """test-package
1.0

/django/contrib/admin/static/admin/css/vendor/select2/LICENSE-SELECT2.md
This is the SELECT2 license text.
/django-5.2.6.dist-info/licenses/LICENSE
This is the primary license text.
/django-5.2.6.dist-info/licenses/LICENSE.python
This is the license for upstream Python.
/django-5.2.6.dist-info/licenses/NOTICE1
This is the first notice file.
/django-5.2.6.dist-info/licenses/NOTICE2
This is the second notice file.
/django-5.2.6.dist-info/licenses/AUTHORS
This is the AUTHORS file.
/django-5.2.6.dist-info/licenses/TESTING
This is a dummy file for testing.

""",
                result,
            )

            args = create_parser().parse_args(
                ["--format=plain-vertical", "--with-license-files", "--with-notice-files", "--with-other-files", "--no-license-path"]
            )
            result = create_output_string(args)
            self.assertEqual(
                """test-package
1.0

This is the SELECT2 license text.
This is the primary license text.
This is the license for upstream Python.
This is the first notice file.
This is the second notice file.
This is the AUTHORS file.
This is a dummy file for testing.

""",
                result,
            )

            args = create_parser().parse_args(["--format=json"])
            result = create_output_string(args)
            self.assertEqual([{"License": "", "Name": "test-package", "Version": "1.0"}], json.loads(result))

            args = create_parser().parse_args(["--format=json", "--with-license-file", "--with-notice-file"])
            result = create_output_string(args)
            self.assertEqual(
                [
                    {
                        "License": "",
                        "LicenseFile": "/django/contrib/admin/static/admin/css/vendor/select2/LICENSE-SELECT2.md",
                        "LicenseText": "This is the SELECT2 license text.",
                        "Name": "test-package",
                        "NoticeFile": "/django-5.2.6.dist-info/licenses/NOTICE1",
                        "NoticeText": "This is the first notice file.",
                        "Version": "1.0",
                    }
                ],
                json.loads(result),
            )

            args = create_parser().parse_args(["--format=json", "--with-license-files", "--with-notice-files", "--with-other-files"])
            result = create_output_string(args)
            self.assertEqual(
                [
                    {
                        "License": "",
                        "LicenseFiles": [
                            "/django/contrib/admin/static/admin/css/vendor/select2/LICENSE-SELECT2.md",
                            "/django-5.2.6.dist-info/licenses/LICENSE",
                            "/django-5.2.6.dist-info/licenses/LICENSE.python",
                        ],
                        "LicenseTexts": ["This is the SELECT2 license text.", "This is the primary license text.", "This is the license for upstream Python."],
                        "Name": "test-package",
                        "NoticeFiles": ["/django-5.2.6.dist-info/licenses/NOTICE1", "/django-5.2.6.dist-info/licenses/NOTICE2"],
                        "NoticeTexts": ["This is the first notice file.", "This is the second notice file."],
                        "OtherFiles": ["/django-5.2.6.dist-info/licenses/AUTHORS", "/django-5.2.6.dist-info/licenses/TESTING"],
                        "OtherTexts": ["This is the AUTHORS file.", "This is a dummy file for testing."],
                        "Version": "1.0",
                    }
                ],
                json.loads(result),
            )



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: Copyright (c) 2018 raimon
# SPDX-FileCopyrightText: Copyright (c) 2025 stefan6419846

import functools
import warnings
from typing import Protocol

from piplicenses.constants import SYSTEM_PACKAGES
from piplicenses.errors import PipLicensesWarning


class SpdxParser(Protocol):
    def __call__(self, expression: str) -> set[str]:  # pragma: no cover
        pass


@functools.lru_cache(maxsize=1)
def _get_spdx_parser() -> SpdxParser:
    """
    Create an SPDX expression parser.

    If the extra "spdx" is not installed, then the parser just returns a set
    with the provided expression as the only element.
    """
    try:
        from license_expression import get_spdx_licensing

    except ImportError:

        def dummy_parser(expression: str) -> set[str]:
            return {expression}

        return dummy_parser

    SYSTEM_PACKAGES.append("license-expression")
    SYSTEM_PACKAGES.append("boolean-py")

    licensing = get_spdx_licensing()

    def parser(expression: str) -> set[str]:
        try:
            result = licensing.validate(expression)
        except Exception:
            # https://github.com/aboutcode-org/license-expression/issues/97
            return {expression}
        if result.errors:
            return {expression}
        parsed = licensing.parse(expression)
        if parsed is None:
            return {expression}
        parsed = parsed.simplify()
        parsed_str = str(parsed)
        if "AND" in parsed_str or "WITH" in parsed_str:
            warnings.warn(
                f"SPDX expressions with 'AND' or 'WITH' are currently not supported. The expression {parsed} is treated as the literal {expression!r}.",
                category=PipLicensesWarning,
                stacklevel=2,
            )
            return {expression}
        return {license_name for license_name in parsed.objects}

    return parser


def _parse_spdx(
    expression: str,
) -> set[str]:
    """Parse a license expression and return a set of licenses."""
    return _get_spdx_parser()(expression)



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: Copyright (c) 2018 raimon
# SPDX-FileCopyrightText: Copyright (c) 2025 stefan6419846

import re
from contextlib import redirect_stderr
from enum import Enum
from io import StringIO
from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryFile
from unittest import TestCase, mock

import tomli_w
from black.mode import auto
from piplicenses_lib import FromArg, __pkgname__

from piplicenses.cli import (
    create_output_string,
    create_parser,
    create_warn_string,
    enum_key_to_value,
    get_output_fields,
    get_sortby,
    load_config_from_file,
    output_colored,
    save_if_needs,
    value_to_enum_key,
)
from piplicenses.constants import TOML_SECTION_NAME
from tests import CaptureOutput, CommandLineTestCase


class CreateOutputStringTestCase(CommandLineTestCase):
    def test_format_html(self) -> None:
        format_html_args = ["--format=html", "--with-authors"]
        args = self.parser.parse_args(format_html_args)
        output_string = create_output_string(args)

        self.assertIn("<table>", output_string)
        self.assertIn("Jukka Lehtosalo &lt;jukka.lehtosalo@iki.fi&gt;", output_string)  # author of "mypy"

    def test_format_json(self) -> None:
        format_json_args = ["--format=json", "--with-authors"]
        args = self.parser.parse_args(format_json_args)
        output_string = create_output_string(args)

        self.assertIn('"Author":', output_string)
        self.assertNotIn('"URL":', output_string)

    def test_format_json_license_manager(self) -> None:
        format_json_args = ["--format=json-license-finder"]
        args = self.parser.parse_args(format_json_args)
        output_string = create_output_string(args)

        self.assertNotIn('"URL":', output_string)
        self.assertIn('"name":', output_string)
        self.assertIn('"version":', output_string)
        self.assertIn('"licenses":', output_string)

    def test_format_csv(self) -> None:
        format_csv_args = ["--format=csv", "--with-authors"]
        args = self.parser.parse_args(format_csv_args)
        output_string = create_output_string(args)

        obtained_header = output_string.split("\n", 1)[0]
        expected_header = '"Name","Version","License","Author"'
        self.assertEqual(obtained_header, expected_header)

    def test_summary(self) -> None:
        summary_args = ["--summary"]
        args = self.parser.parse_args(summary_args)
        output_string = create_output_string(args)

        self.assertIn("Count", output_string)
        self.assertNotIn("Name", output_string)

        warn_string = create_warn_string(args)
        self.assertEqual("", warn_string)

    def test_format_plain_vertical(self) -> None:
        format_plain_args = ["--format=plain-vertical", "--from=classifier"]
        args = self.parser.parse_args(format_plain_args)
        output_string = create_output_string(args)
        self.assertIsNotNone(re.search(r"pep8-naming\n\d+\.\d+\.\d+\nMIT License\n", output_string), output_string)


class ParserTestCase(TestCase):
    def _assert_system_exit_and_get_stderr(self, parameters: list[str]) -> str:
        parser = create_parser()
        stderr = StringIO()
        with self.assertRaises(expected_exception=SystemExit), redirect_stderr(stderr):
            parser.parse_args(parameters)
        return stderr.getvalue()

    def test_verify_args(self) -> None:
        # --with-license-file missing
        stderr = self._assert_system_exit_and_get_stderr(["--no-license-path"])
        for arg in ("--no-license-path", "--with-license-file"):
            self.assertIn(arg, stderr)

        stderr = self._assert_system_exit_and_get_stderr(["--with-notice-file"])
        for arg in ("--with-notice-file", "--with-license-file"):
            self.assertIn(arg, stderr)

        # --filter-strings missing
        stderr = self._assert_system_exit_and_get_stderr(["--filter-code-page=utf8"])
        for arg in ("--filter-code-page", "--filter-strings"):
            self.assertIn(arg, stderr)

        # invalid code-page
        stderr = self._assert_system_exit_and_get_stderr(["--filter-strings", "--filter-code-page=XX"])
        for arg in ("invalid code", "--filter-code-page"):
            self.assertIn(arg, stderr)

    def test_pyproject_toml_args_parsed_correctly(self) -> None:
        # We test that parameters of different types are deserialized correctly.
        pyproject_conf = {
            "tool": {
                TOML_SECTION_NAME: {
                    # choices_from_enum
                    "from": "classifier",
                    # bool
                    "summary": True,
                    # list[str]
                    "ignore-packages": ["package1", "package2"],
                    # str
                    "fail-on": "LIC1;LIC2",
                }
            }
        }

        toml_str = tomli_w.dumps(pyproject_conf)

        # Create a temporary file and write the TOML string to it.
        with NamedTemporaryFile(suffix=".toml", delete=False) as temp_file:
            temp_file.write(toml_str.encode("utf-8"))
            temp_file.seek(0)

            parser = create_parser(temp_file.name)
            args = parser.parse_args([])

            tool_conf = pyproject_conf["tool"][TOML_SECTION_NAME]

            # Assert values are correctly parsed from toml.
            self.assertEqual(FromArg.CLASSIFIER, args.from_)
            self.assertEqual(tool_conf["summary"], args.summary)
            self.assertEqual(tool_conf["ignore-packages"], args.ignore_packages)
            self.assertEqual(tool_conf["fail-on"], args.fail_on)

            # Assert args are rewritable using cli.
            args = parser.parse_args(["--from=meta"])

            self.assertEqual(FromArg.META, args.from_)

            # All other values are parsed from toml.
            self.assertEqual(tool_conf["summary"], args.summary)
            self.assertEqual(tool_conf["ignore-packages"], args.ignore_packages)
            self.assertEqual(tool_conf["fail-on"], args.fail_on)


class LoadConfigFromFileTestCase(TestCase):
    def test_load_non_existent_file(self) -> None:
        with NamedTemporaryFile() as fd:
            pass
        self.assertDictEqual({}, load_config_from_file(fd.name))


class EnumTestCase(TestCase):
    def test_functions(self) -> None:
        class TestEnum(Enum):
            PLAIN = P = auto()
            JSON_LICENSE_FINDER = JLF = auto()

        self.assertEqual(TestEnum.PLAIN, TestEnum.P)
        self.assertEqual(TestEnum.JSON_LICENSE_FINDER, getattr(TestEnum, value_to_enum_key("jlf")))

        self.assertEqual("JLF", value_to_enum_key("jlf"))
        self.assertEqual("JSON_LICENSE_FINDER", value_to_enum_key("json-license-finder"))

        self.assertEqual("json-license-finder", enum_key_to_value(TestEnum.JSON_LICENSE_FINDER))
        self.assertEqual("plain", enum_key_to_value(TestEnum.PLAIN))


class SaveIfNeedsTestCase(TestCase):
    def test_output_file_success(self) -> None:
        def mocked_open(*args, **kwargs):
            return TemporaryFile("w")

        with mock.patch("piplicenses.cli.open", mocked_open), mock.patch("sys.exit"), CaptureOutput() as captured:
            save_if_needs("/foo/bar.txt", "license list")

        self.assertIn("created path: ", captured.stdout)
        self.assertEqual("", captured.stderr)

    def test_output_file_error(self) -> None:
        def mocked_open(*args, **kwargs):
            raise OSError

        with mock.patch("piplicenses.cli.open", mocked_open), mock.patch("sys.exit"), CaptureOutput() as captured:
            save_if_needs("/foo/bar.txt", "license list")

        self.assertEqual("", captured.stdout)
        self.assertIn("check path: ", captured.stderr)

    def test_output_file_none(self) -> None:
        with mock.patch("sys.exit"), CaptureOutput() as captured:
            save_if_needs(None, "license list")

        # stdout and stderr are expected not to be called.
        self.assertEqual("", captured.stdout)
        self.assertEqual("", captured.stderr)

    def test_output_file_content(self) -> None:
        with NamedTemporaryFile() as fd, mock.patch("sys.exit"), CaptureOutput() as captured:
            fd.close()

            save_if_needs(fd.name, "Hello World!")
            self.assertEqual("Hello World!\n", Path(fd.name).read_text())

            save_if_needs(fd.name, "Hello World!\n")
            self.assertEqual("Hello World!\n", Path(fd.name).read_text())

        self.assertEqual(f"created path: {fd.name}\n" * 2, captured.stdout)
        self.assertEqual("", captured.stderr)


class OutputColoredTestCase(TestCase):

    def test_output_colored_normal(self) -> None:
        color_code = "32"
        text = __pkgname__
        actual = output_colored(color_code, text)

        self.assertTrue(actual.startswith("\033[32"))
        self.assertIn(text, actual)
        self.assertTrue(actual.endswith("\033[0m"))

    def test_output_colored_bold(self) -> None:
        color_code = "32"
        text = __pkgname__
        actual = output_colored(color_code, text, is_bold=True)

        self.assertTrue(actual.startswith("\033[1;32"))
        self.assertIn(text, actual)
        self.assertTrue(actual.endswith("\033[0m"))


class CreateWarnStringTestCase(CommandLineTestCase):
    def test_summary_warning(self) -> None:
        summary_args = ["--summary", "--with-authors"]
        args = self.parser.parse_args(summary_args)

        warn_string = create_warn_string(args)
        self.assertIn(
            "using --with-authors and --with-urls will be ignored.",
            warn_string,
        )

        summary_args = ["--summary", "--with-urls"]
        args = self.parser.parse_args(summary_args)

        warn_string = create_warn_string(args)
        self.assertIn(
            "using --with-authors and --with-urls will be ignored.",
            warn_string,
        )

    def test_with_license_files_format_warning(self) -> None:
        args = self.parser.parse_args(["--format=html", "--with-license-files"])
        warn_string = create_warn_string(args)
        self.assertIn("Ignoring request to output multiple files due to unsupported output format.", warn_string)

        args = self.parser.parse_args(["--format=json", "--with-license-files"])
        warn_string = create_warn_string(args)
        self.assertNotIn("Ignoring request to output multiple files due to unsupported output format.", warn_string)


class GetSortbyTestCase(CommandLineTestCase):
    def test_summary_sort_by_count(self) -> None:
        summary_args = ["--summary", "--order=count"]
        args = self.parser.parse_args(summary_args)

        sortby = get_sortby(args)
        self.assertEqual("Count", sortby)

    def test_summary_sort_by_name(self) -> None:
        summary_args = ["--summary", "--order=name"]
        args = self.parser.parse_args(summary_args)

        sortby = get_sortby(args)
        self.assertEqual("License", sortby)

    def test_order_name(self) -> None:
        order_name_args = ["--order=name"]
        args = self.parser.parse_args(order_name_args)

        sortby = get_sortby(args)
        self.assertEqual("Name", sortby)

    def test_order_license(self) -> None:
        order_license_args = ["--order=license"]
        args = self.parser.parse_args(order_license_args)

        sortby = get_sortby(args)
        self.assertEqual("License", sortby)

    def test_order_author(self) -> None:
        order_author_args = ["--order=author", "--with-authors"]
        args = self.parser.parse_args(order_author_args)

        sortby = get_sortby(args)
        self.assertEqual("Author", sortby)

    def test_order_maintainer(self) -> None:
        order_maintainer_args = ["--order=maintainer", "--with-maintainers"]
        args = self.parser.parse_args(order_maintainer_args)

        sortby = get_sortby(args)
        self.assertEqual("Maintainer", sortby)

    def test_order_url(self) -> None:
        order_url_args = ["--order=url", "--with-urls"]
        args = self.parser.parse_args(order_url_args)

        sortby = get_sortby(args)
        self.assertEqual("URL", sortby)

    def test_order_url_no_effect(self) -> None:
        order_url_args = ["--order=url"]
        args = self.parser.parse_args(order_url_args)

        sortby = get_sortby(args)
        self.assertEqual("Name", sortby)


class GetOutputFieldsTestCase(CommandLineTestCase):
    def test_with_license_files(self) -> None:
        for format_string in ["json", "plain-vertical"]:
            with self.subTest(format_string=format_string):
                args = self.parser.parse_args(["--with-license-files", f"--format={format_string}", "--with-notice-files", "--with-other-files"])
                fields = get_output_fields(args)
                self.assertEqual(
                    ["Name", "Version", "License", "LicenseFiles", "LicenseTexts", "NoticeFiles", "NoticeTexts", "OtherFiles", "OtherTexts"], fields
                )

        for format_string in ["plain", "csv", "html"]:
            with self.subTest(format_string=format_string):
                args = self.parser.parse_args(["--with-license-files", f"--format={format_string}", "--with-notice-files", "--with-other-files"])
                fields = get_output_fields(args)
                self.assertEqual(["Name", "Version", "License"], fields)

    def test_files_singular_plural(self) -> None:
        args = self.parser.parse_args(
            [
                "--with-license-file",
                "--format=json",
                "--with-notice-file",
                "--with-other-files",
            ]
        )
        fields = get_output_fields(args)
        self.assertEqual(["Name", "Version", "License", "LicenseFile", "LicenseText", "NoticeFile", "NoticeText", "OtherFiles", "OtherTexts"], fields)

        args = self.parser.parse_args(
            [
                "--with-license-files",
                "--format=json",
                "--with-notice-files",
                "--with-other-files",
            ]
        )
        fields = get_output_fields(args)
        self.assertEqual(["Name", "Version", "License", "LicenseFiles", "LicenseTexts", "NoticeFiles", "NoticeTexts", "OtherFiles", "OtherTexts"], fields)

    def test_no_license_path(self) -> None:
        args = self.parser.parse_args(
            [
                "--with-license-files",
                "--format=json",
                "--with-notice-files",
                "--with-other-files",
            ]
        )
        fields = get_output_fields(args)
        self.assertEqual(["Name", "Version", "License", "LicenseFiles", "LicenseTexts", "NoticeFiles", "NoticeTexts", "OtherFiles", "OtherTexts"], fields)

        args = self.parser.parse_args(
            [
                "--with-license-files",
                "--format=json",
                "--with-notice-files",
                "--with-other-files",
                "--no-license-path",
            ]
        )
        fields = get_output_fields(args)
        self.assertEqual(["Name", "Version", "License", "LicenseTexts", "NoticeTexts", "OtherTexts"], fields)



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: Copyright (c) 2018 raimon
# SPDX-FileCopyrightText: Copyright (c) 2025 stefan6419846

from __future__ import annotations

from collections import Counter
from typing import TYPE_CHECKING, cast

from piplicenses_lib import LICENSE_UNKNOWN
from prettytable import HRuleStyle, PrettyTable

from piplicenses.collector import get_packages
from piplicenses.constants import _MULTI_VALUE_KEYS, DEFAULT_OUTPUT_FIELDS, FIELDS_TO_METADATA_KEYS, SUMMARY_FIELD_NAMES

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Iterator, Sequence

    from prettytable import RowType

    from piplicenses.cli import CustomNamespace


def _handle_multiple_value_field(key: str, value: Iterator[str]) -> str | list[str]:
    if key.endswith("s"):
        return list(value) or ["UNKNOWN"]
    return cast(str, next(value, LICENSE_UNKNOWN))


def create_licenses_table(
    args: CustomNamespace,
    output_fields: Sequence[str] = DEFAULT_OUTPUT_FIELDS,
) -> PrettyTable:
    table = factory_styled_table_with_args(args, output_fields)

    for pkg in get_packages(args):
        row: list[str | list[str]] = []
        for field in output_fields:
            if field == "License":
                license_set = pkg.license_names
                license_str = "; ".join(sorted(license_set))
                row.append(license_str)
            elif field == "License-Classifier":
                row.append("; ".join(sorted(pkg.license_classifiers)) or LICENSE_UNKNOWN)
            elif hasattr(pkg, field.lower()):
                row.append(cast(str, getattr(pkg, field.lower())))
            else:
                value = getattr(pkg, FIELDS_TO_METADATA_KEYS[field])
                if field in _MULTI_VALUE_KEYS:
                    row.append(_handle_multiple_value_field(field, value))
                else:
                    row.append(cast(str, value))
        table.add_row(row)

    return table


def create_summary_table(args: CustomNamespace) -> PrettyTable:
    counts = Counter("; ".join(sorted(pkg.license_names)) for pkg in get_packages(args))

    table = factory_styled_table_with_args(args, SUMMARY_FIELD_NAMES)
    for license_, count in counts.items():
        table.add_row([count, license_])
    return table


class JsonPrettyTable(PrettyTable):
    """PrettyTable-like class exporting to JSON"""

    def format_row(self, row: RowType) -> dict[str, str | list[str]]:
        resrow: dict[str, str | list[str]] = {}
        for field, value in zip(self._field_names, row):
            resrow[field] = value

        return resrow

    def get_string(self, **kwargs: str | list[str]) -> str:
        # import included here in order to limit dependencies
        # if not interested in JSON output,
        # then the dependency is not required
        import json

        options = self._get_options(kwargs)
        rows = self._get_rows(options)
        lines = [self.format_row(row) for row in rows]
        return json.dumps(lines, indent=2, sort_keys=True)


class JsonLicenseFinderTable(JsonPrettyTable):
    def format_row(self, row: RowType) -> dict[str, str | list[str]]:
        resrow: dict[str, str | list[str]] = {}
        for field, value in zip(self._field_names, row):
            if field == "Name":
                resrow["name"] = value

            if field == "Version":
                resrow["version"] = value

            if field == "License":
                resrow["licenses"] = [value]

        return resrow

    def get_string(self, **kwargs: str | list[str]) -> str:
        # import included here in order to limit dependencies
        # if not interested in JSON output,
        # then the dependency is not required
        import json

        options = self._get_options(kwargs)
        rows = self._get_rows(options)
        lines = [self.format_row(row) for row in rows]
        return json.dumps(lines, sort_keys=True)


class CSVPrettyTable(PrettyTable):
    """PrettyTable-like class exporting to CSV"""

    def get_string(self, **kwargs: str | list[str]) -> str:
        def esc_quotes(val: bytes | str) -> str:
            """
            Meta-escaping double quotes
            https://tools.ietf.org/html/rfc4180
            """
            try:
                return cast(str, val).replace('"', '""')
            except UnicodeDecodeError:  # pragma: no cover
                return cast(bytes, val).decode("utf-8").replace('"', '""')
            except UnicodeEncodeError:  # pragma: no cover
                return str(cast(str, val).encode("unicode_escape").replace('"', '""'))  # type: ignore[arg-type]

        options = self._get_options(kwargs)
        rows = self._get_rows(options)
        formatted_rows = self._format_rows(rows)

        lines: list[str] = []
        formatted_header = ",".join([f'"{esc_quotes(val)}"' for val in self._field_names])
        lines.append(formatted_header)
        lines.extend([",".join([f'"{esc_quotes(val)}"' for val in row]) for row in formatted_rows])
        return "\n".join(lines)


class PlainVerticalTable(PrettyTable):
    """PrettyTable for outputting to a simple non-column based style.

    When used with --with-license-file, this style is similar to the default
    style generated from Angular CLI's --extractLicenses flag.
    """

    def get_string(self, **kwargs: str | list[str]) -> str:
        options = self._get_options(kwargs)
        rows = self._get_rows(options)
        show_paths = "LicenseFiles" in kwargs["fields"]

        output = ""
        for row in rows:
            index = 0
            while index < len(row):
                v = row[index]
                if isinstance(v, list):
                    if show_paths:
                        for first_entry, second_entry in zip(v, row[index + 1]):
                            output += "{}\n{}\n".format(first_entry, second_entry)
                        index += 1
                    else:
                        for entry in v:
                            output += "{}\n".format(entry)
                else:
                    output += "{}\n".format(v)
                index += 1
            output += "\n"

        return output


def factory_styled_table_with_args(
    args: CustomNamespace,
    output_fields: Sequence[str] = DEFAULT_OUTPUT_FIELDS,
) -> PrettyTable:
    from piplicenses.cli import FormatArg

    table = PrettyTable()
    table.field_names = output_fields  # type: ignore[assignment]
    table.align = "l"  # type: ignore[assignment]
    table.border = args.format_ in {
        FormatArg.MARKDOWN,
        FormatArg.RST,
        FormatArg.CONFLUENCE,
        FormatArg.JSON,
    }
    table.header = True

    if args.format_ == FormatArg.MARKDOWN:
        table.junction_char = "|"
        table.hrules = HRuleStyle.HEADER
    elif args.format_ == FormatArg.RST:
        table.junction_char = "+"
        table.hrules = HRuleStyle.ALL
    elif args.format_ == FormatArg.CONFLUENCE:
        table.junction_char = "|"
        table.hrules = HRuleStyle.NONE
    elif args.format_ == FormatArg.JSON:
        table = JsonPrettyTable(table.field_names)
    elif args.format_ == FormatArg.JSON_LICENSE_FINDER:
        table = JsonLicenseFinderTable(table.field_names)
    elif args.format_ == FormatArg.CSV:
        table = CSVPrettyTable(table.field_names)
    elif args.format_ == FormatArg.PLAIN_VERTICAL:
        table = PlainVerticalTable(table.field_names)

    return table
